/**
 * These implementations mirror the Bitcoin Script based implementations used.
 * They are necessary to ensure that we produce the same results as the oracles.
 */
export declare const precision = 64;
export declare const scale: number;
export declare const bigIntScale: bigint;
export declare const e: bigint;
/**
 * Calculates the binary exponent of x using the binary fraction method.
 * Accepts and returns scaled by 2**64 (64-bit fixed-point number).
 * Adapted from https://github.com/paulrberg/prb-math
 */
export declare function exp2(x: bigint): bigint;
/**
 * Accepts and returns scaled by 2**64 (64-bit fixed-point number).
 */
export declare function exp(x: bigint): bigint;
/**
 * Finds the zero-based index of the first one in the binary representation of x.
 * Accepts numbers scaled by 2**64 (64-bit fixed-point number).
 * Adapted from https://github.com/paulrberg/prb-math
 */
export declare function mostSignificantBit(x: bigint): number;
/**
 * Calculates the binary logarithm of x.
 * Accepts and returns scaled by 2**64 (64-bit fixed-point number).
 * Only works for x greater then 1 << 64 (log2(1)).
 * Adapted from https://github.com/paulrberg/prb-math
 */
export declare function log2(x: bigint): bigint;
export declare function log(x: bigint): bigint;
export declare function log10(x: bigint): bigint;
